import numpy as np
import numpy.ma as ma

import pandas as pd
from pandas.tseries.frequencies import to_offset


"""
Basic QA/QC tools.

"""


def has_time_gaps(times, freq):
    """
    Check for gaps in a series time-stamp `times`. The `freq` can be a string
    or a pandas offset object.  Note the `freq` cannot be an ambiguous offset
    (like week, months, etc), in those case reduce it to the smallest
    unambiguous unit (i.e.: 1 month -> 30 days).

    Example
    -------
    >>> import numpy as np
    >>> import pandas as pd
    >>> times = pd.date_range('1980-01-19', periods=48, freq='1H')
    >>> has_time_gaps(times, freq='6min')
    True
    >>> has_time_gaps(times, freq='1H')
    False

    """
    freq = to_offset(freq)
    if hasattr(freq, 'delta'):
        times = pd.DatetimeIndex(times)
    else:
        raise ValueError('Cannot interpret freq {!r} delta'.format(freq))
    return (np.diff(times) > freq.delta.to_timedelta64()).any()


def is_monotonically_increasing(series):
    """
    Check if a given list or array is monotonically increasing.

    Examples
    --------
    >>> import pandas as pd
    >>> times = pd.date_range('1980-01-19', periods=10)
    >>> all(is_monotonically_increasing(times))
    True
    >>> import numpy as np
    >>> all(is_monotonically_increasing(np.r_[times[-2:-1], times]))
    False
    """
    return [x < y for x, y in zip(series, series[1:])]


def is_flatline(series, reps=10, eps=None):
    """
    Check for consecutively repeated values (`reps`) in `series` within a
    tolerance `eps`.

    Examples
    --------
    >>> import numpy as np
    >>> series = np.r_[np.random.rand(10), [10]*15, np.random.rand(10)]
    >>> is_flatline(series, reps=10)
    array([False, False, False, False, False, False, False, False, False,
           False,  True,  True,  True,  True,  True,  True,  True,  True,
            True,  True,  True,  True,  True,  True,  True, False, False,
           False, False, False, False, False, False, False, False])

    """
    series = np.asanyarray(series)

    if not eps:
        eps = np.finfo(float).eps

    if reps < 2:
        reps = 2

    mask = np.zeros_like(series, dtype='bool')

    flatline = 1
    for k, current in enumerate(series):
        if np.abs(series[k-1] - current) < eps:
            flatline += 1
        else:
            if flatline >= reps:
                mask[k-flatline:k] = True
            flatline = 1
    return mask


def is_spike(series, window_size=3, threshold=3, scale=True):
    """
    Flags spikes in an array-like object using a median filter of `window_size`
    and a `threshold` for the median difference.  If `scale=False` the
    differences are not scale by the data standard deviation and the masking
    is "aggressive."

    Examples
    --------
    >>> import pandas as pd
    >>> series = [33.43, 33.45, 34.45, 90.0, 35.67, 34.9, 43.5, 34.6, 33.7]
    >>> series = pd.Series(series, index=pd.date_range('1980-01-19',
    ...                    periods=len(series)))
    >>> series[is_spike(series, window_size=3, threshold=3, scale=False)]
    1980-01-22    90.0
    1980-01-25    43.5
    dtype: float64
    >>> series[is_spike(series, window_size=3, threshold=3, scale=True)]
    1980-01-22    90.0
    Freq: D, dtype: float64

    """
    # bfill+ffil needs a series and won't affect the median.
    series = pd.Series(series)
    roll = series.rolling(window=window_size, center=True)
    medians = roll.median().fillna(method='bfill').fillna(method='ffill')
    difference = np.abs(series - medians).values
    if scale:
        return difference > (threshold*difference.std())
    return difference > threshold


def first_difference(series, quantile=0.75):
    """
    Flags spikes by using a first difference approach
    The differences are compared against the `quantile` (default is 0.75).
    Note that first difference might remove a "good" data point right after
    a high diff is found.

    Examples
    --------
    >>> import pandas as pd
    >>> series = [33.43, 33.45, 34.45, 90.0, 35.67, 34.9, 43.5, 34.6, 33.7]
    >>> series = pd.Series(series, index=pd.date_range('1980-01-19',
    ...                    periods=len(series)))
    >>> series[first_difference(series, quantile=0.75)]
    1980-01-22    90.00
    1980-01-23    35.67
    Freq: D, dtype: float64

    """
    series = pd.Series(series)
    abs_diffs = series.diff().abs().fillna(method='bfill').fillna(method='ffill')  # noqa
    return abs_diffs > abs_diffs.quantile(quantile)


def threshold_series(series, vmin=None, vmax=None):
    """
    Threshold an series by flagging with NaN values below `vmin` and above
    `vmax`.

    Examples
    --------
    >>> series = [0.1, 20, 30, 35.5, 34.9, 43.5, 34.6, 40]
    >>> threshold_series(series, vmin=30, vmax=40)
    masked_array(data=[--, --, 30.0, 35.5, 34.9, --, 34.6, 40.0],
                 mask=[ True,  True, False, False, False,  True, False, False],
           fill_value=1e+20)
    >>> import pandas as pd
    >>> series = pd.Series(series, index=pd.date_range('1980-01-19',
    ...                    periods=len(series)))
    >>> threshold_series(series, vmin=30, vmax=40)
    1980-01-19     NaN
    1980-01-20     NaN
    1980-01-21    30.0
    1980-01-22    35.5
    1980-01-23    34.9
    1980-01-24     NaN
    1980-01-25    34.6
    1980-01-26    40.0
    Freq: D, dtype: float64

    """
    if not vmin:
        vmin = min(series)
    if not vmax:
        vmax = max(series)

    masked = ma.masked_outside(series, vmin, vmax)
    if masked.mask.any():
        if isinstance(series, pd.Series):
            series[masked.mask] = np.NaN
            return series
    return masked


def filter_spikes(series, window_size=3, threshold=3, scale=True):
    """
    Filter an array-like object using a median filter and a `threshold`
    for the median difference.

    Examples
    --------
    >>> import pandas as pd
    >>> series = [33.43, 33.45, 34.45, 90.0, 35.67, 34.9, 43.5, 34.6, 33.7]
    >>> series = pd.Series(series, index=pd.date_range('1980-01-19',
    ...                    periods=len(series)))
    >>> filter_spikes(series)
    1980-01-19    33.43
    1980-01-20    33.45
    1980-01-21    34.45
    1980-01-22      NaN
    1980-01-23    35.67
    1980-01-24    34.90
    1980-01-25    43.50
    1980-01-26    34.60
    1980-01-27    33.70
    Freq: D, dtype: float64

    """
    outlier_idx = is_spike(series, window_size=window_size,
                           threshold=threshold, scale=scale)
    if not isinstance(series, pd.Series):
        series = np.asanyarray(series)
    series[outlier_idx] = np.NaN
    return series


def _high_pass(data, alpha=0.5):
    """
    Runs a high pass RC filter over the given data.

    Parameters
    ----------
    data : array_like

    alpha : float
            Smoothing factor between 0.0 (exclusive) and 1.0 (inclusive).
            A lower value means more filtering.  A value of 1.0 equals no
            filtering. Defaults is 0.5.

    Returns
    -------
    hpf : array_like
          Filtered data.

    Based on http://en.wikipedia.org/wiki/High-pass_filter

    """
    mean = data.mean()
    data = data - mean
    hpf = data.copy()
    for k in range(1, len(data)):
        hpf[k] = alpha * hpf[k-1] + alpha * (data[k] - data[k-1])
    return hpf + mean


def tukey53H(series, k=1.5):
    """
    Flags suspicious spikes values in `series` using Tukey 53H criteria.

    References
    ----------
    .. [1] Goring, Derek G., and Vladimir I. Nikora. "Despiking acoustic
    Doppler velocimeter data." Journal of Hydraulic Engineering 128.1 (2002):
    117-126.  http://dx.doi.org/10.1061/(ASCE)0733-9429(2002)128:1(117)

    Examples
    --------
    >>> import pandas as pd
    >>> series = [33.43, 33.45, 34.45, 90.0, 35.67, 34.9, 43.5, 34.6, 33.7]
    >>> series = pd.Series(series, index=pd.date_range('1980-01-19',
    ...                    periods=len(series)))
    >>> series[tukey53H(series, k=1.5)]
    1980-01-22    90.0
    Freq: D, dtype: float64

    """
    series = np.asanyarray(series)

    series = _high_pass(series, 0.99)
    series = series - series.mean()

    N = len(series)
    stddev = series.std()

    u1 = np.zeros_like(series)
    for n in range(N-4):
        if series[n:n+5].any():
            u1[n+2] = np.median(series[n:n+5])

    u2 = np.zeros_like(series)
    for n in range(N-2):
        if u1[n:n+3].any():
            u2[n+1] = np.median(u1[n:n+3])

    u3 = np.zeros_like(series)
    u3[1:-1] = 0.25*(u2[:-2] + 2*u2[1:-1] + u2[2:])

    delta = np.abs(series-u3)

    return delta > k*stddev



from datetime import datetime

import cf_units

import cftime

import iris
from iris import Constraint
from iris.coords import AuxCoord, DimCoord
from iris.cube import CubeList
from iris.exceptions import CoordinateMultiDimError, CoordinateNotFoundError

import numpy as np

import pandas as pd
from pandas.core.indexes.datetimes import DatetimeIndex

from scipy.spatial import cKDTree as KDTree


"""
Tools to manipulate iris cubes.

"""


def is_model(cube):
    """
    Heuristic way to find if a cube data is `modelResult` or not.
    WARNING: This function may return False positives and False
    negatives!!!

    Examples
    --------
    >>> import iris
    >>> iris.FUTURE.netcdf_promote = True
    >>> url = ("http://crow.marine.usf.edu:8080/thredds/dodsC/"
    ...        "FVCOM-Nowcast-Agg.nc")
    >>> cubes = iris.load_raw(url, 'sea_surface_height_above_geoid')
    >>> [is_model(cube) for cube in cubes]
    [True]
    >>> url = ("http://thredds.cdip.ucsd.edu/thredds/dodsC/cdip/archive/"
    ...        "043p1/043p1_d17.nc")
    >>> cubes = iris.load_raw(url, 'sea_surface_temperature')
    >>> [is_model(cube) for cube in cubes]
    [False]

    """
    # First criteria (Strong): "forecast" word in the time coord.
    try:
        coords = cube.coords(axis='T')
        for coord in coords:
            if 'forecast' in coord.name():
                return True
    except CoordinateNotFoundError:
        pass
    # Second criteria (Strong): `UGRID` cubes are models.
    conventions = cube.attributes.get('Conventions', 'None')
    if 'UGRID' in conventions.upper():
        return True
    # Third criteria (Strong): dimensionless coords are present.
    try:
        coords = cube.coords(axis='Z')
        for coord in coords:
            if 'ocean_' in coord.name():
                return True
    except CoordinateNotFoundError:
        pass
    # Forth criteria (weak): Assumes that all "GRID" attribute are models.
    cdm_data_type = cube.attributes.get('cdm_data_type', 'None')
    feature_type = cube.attributes.get('featureType', 'None')
    source = cube.attributes.get('source', 'None')
    if cdm_data_type.upper() == 'GRID' or feature_type.upper() == 'GRID':
        if 'AVHRR' not in source:
            return True
    return False


def z_coord(cube):
    """
    Return the canonical vertical coordinate.

    Examples
    --------
    >>> import iris
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> zvar = z_coord(cube)
    >>> zvar.name()
    'ocean_s_coordinate_g2'
    >>> cube.coord_dims(zvar)
    (1,)

    """
    # http://cfconventions.org/cf-conventions/v1.6.0/cf-conventions.html#dimensionless-v-coord
    dimensionless = [
        'atmosphere_hybrid_height_coordinate',
        'atmosphere_hybrid_sigma_pressure_coordinate',
        'atmosphere_sigma_coordinate',
        'atmosphere_sleve_coordinate',
        'ocean_s_coordinate',
        'ocean_s_coordinate_g1',
        'ocean_s_coordinate_g2',
        'ocean_sigma_coordinate',
        'ocean_sigma_z_coordinate'
        ]

    # Setting `dim_coords=True` to avoid triggering the download
    # of the derived `z`. That usually throws a Memory error as it
    # varies with `t`, `x`, `y`, and `z`.
    zvars = cube.coords(axis='Z', dim_coords=True)
    if not zvars:
        zvars = cube.coords(axis='altitude', dim_coords=True)

    # Could not find a dimension coordinate.
    if not zvars:
        zvars = [coord for coord in cube.coords(axis='Z')
                 if coord.name() in dimensionless]

    if not zvars:
        raise ValueError(f'Could not find the vertical coordinate!')

    if len(zvars) > 1:
        raise ValueError(f'Found more than one vertical coordinate: {zvars}')
    else:
        zvar = zvars[0]

    return zvar


def _get_surface_idx(cube):
    """
    Return the `cube` index for the surface layer of for any model grid
    (rgrid, ugrid, sgrid), and any non-dimensional coordinate.

    """
    z = z_coord(cube)
    if not z:
        raise ValueError(f'Cannot find the surface for cube {repr(cube)}')
    else:
        if np.argmin(z.shape) == 0 and z.ndim == 2:
            points = z[:, 0].points
        elif np.argmin(z.shape) == 1 and z.ndim == 2:
            points = z[0, :].points
        else:
            points = z.points
        positive = z.attributes.get('positive', None)
        if positive == 'up':
            idx = np.unique(points.argmax(axis=0))[0]
        else:
            idx = np.unique(points.argmin(axis=0))[0]
        return idx


def get_surface(cube, conventions='None'):
    """
    Work around `iris.cube.Cube.slices` error:
    The requested coordinates are not orthogonal.

    Examples
    --------
    >>> import iris
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> cube.ndim == 4
    True
    >>> get_surface(cube).ndim == 3
    True

    """
    conventions = cube.attributes.get('Conventions', conventions)

    # Short-circuit if cube does not have a z-axis.
    if 'UGRID' not in conventions.upper() and cube.ndim == 3:
        return cube
    if 'UGRID' in conventions.upper() and cube.ndim == 2:
        return cube

    idx = _get_surface_idx(cube)
    if cube.ndim == 4 or 'UGRID' in conventions.upper():
        return cube[:, int(idx), ...]
    elif cube.ndim == 3 and 'UGRID' not in conventions.upper():
        return cube[int(idx), ...]
    else:
        raise ValueError(f'Cannot find the surface for cube {repr(cube)}')


def time_coord(cube):
    """
    Return the variable attached to time axis.

    Examples
    --------
    >>> import iris
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> timevar = time_coord(cube)
    >>> timevar.name()  # What is the time coordinate named?
    'time'
    >>> cube.coord_dims(timevar)  # Is it the zeroth coordinate?
    (0,)

    """

    timevars = cube.coords(axis='T', dim_coords=True)
    if not timevars:
        timevars = [coord for coord in cube.dim_coords if 'time' in coord.name()]  # noqa
        if not timevars:
            ValueError(f'Could not find "time" in {repr(cube.dim_coords)}')
    if len(timevars) != 1:
        raise ValueError('Found more than one time coordinates!')
    timevar = timevars[0]
    return timevar


def time_near(cube, datetime_obj):
    """
    Return the nearest index to a `datetime_obj`.

    Examples
    --------
    >>> import iris
    >>> from datetime import datetime
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> isinstance(time_near(cube, datetime.utcnow()), np.integer)
    True

    """
    timevar = time_coord(cube)
    try:
        time = timevar.units.date2num(datetime_obj)
        idx = timevar.nearest_neighbour_index(time)
    except IndexError:
        idx = -1
    return idx


def time_slice(cube, start, stop=None):
    """
    Slice time by indexes using a nearest criteria.
    NOTE: Assumes time is the first dimension!

    Examples
    --------
    >>> import iris
    >>> from datetime import datetime, timedelta
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> stop = datetime.utcnow()
    >>> start = stop - timedelta(days=7)
    >>> time_slice(cube, start, stop).shape[0] < cube.shape[0]
    True

    """
    istart = time_near(cube, start)
    if stop:
        istop = time_near(cube, stop)
        if istart == istop:
            raise ValueError(
                f'istart must be different from istop! '
                f'Got istart {istart} and '
                f' istop {istop}'
                )
        return cube[istart:istop, ...]
    else:
        return cube[istart, ...]


def _minmax(v):
    return np.min(v), np.max(v)


def _get_indices(cube, bbox):
    """
    Get the 4 corner indices of a `cube` given a `bbox`.

    Examples
    --------
    >>> import iris
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> bbox = [-87.40, 24.25, -74.70, 36.70]
    >>> idxs = _get_indices(cube, bbox)
    >>> [isinstance(idx, np.integer) for idx in idxs]
    [True, True, True, True]
    >>> idxs
    (180, 335, 119, 250)

    """
    from oceans.ocfis import wrap_lon180
    lons = cube.coord('longitude').points
    lats = cube.coord('latitude').points
    lons = wrap_lon180(lons)

    inregion = np.logical_and(np.logical_and(lons > bbox[0],
                                             lons < bbox[2]),
                              np.logical_and(lats > bbox[1],
                                             lats < bbox[3]))
    region_inds = np.where(inregion)
    imin, imax = _minmax(region_inds[0])
    jmin, jmax = _minmax(region_inds[1])
    return imin, imax+1, jmin, jmax+1


def bbox_extract_2Dcoords(cube, bbox):
    """
    Extract a sub-set of a cube inside a lon, lat bounding box
    bbox = [lon_min lon_max lat_min lat_max].
    NOTE: This is a work around too subset an iris cube that has
    2D lon, lat coords.

    Examples
    --------
    >>> import iris
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> bbox = [-87.40, 24.25, -74.70, 36.70]
    >>> new_cube = bbox_extract_2Dcoords(cube, bbox)
    >>> cube.shape != new_cube.shape
    True

    """
    imin, imax, jmin, jmax = _get_indices(cube, bbox)
    return cube[..., imin:imax, jmin:jmax]


def bbox_extract_1Dcoords(cube, bbox):
    """
    Same as bbox_extract_2Dcoords but for 1D coords.

    Examples
    --------
    >>> import iris
    >>> url = "http://oos.soest.hawaii.edu/thredds/dodsC/pacioos/hycom/global"
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> bbox = [272.6, 24.25, 285.3, 36.70]
    >>> new_cube = bbox_extract_1Dcoords(cube, bbox)
    >>> cube.shape != new_cube.shape
    True

    """
    lat = Constraint(latitude=lambda cell: bbox[1] <= cell <= bbox[3])
    lon = Constraint(longitude=lambda cell: bbox[0] <= cell <= bbox[2])
    cube = cube.extract(lon & lat)
    return cube


def subset(cube, bbox):
    """
    Subsets cube with 1D or 2D lon, lat coords.
    Using `intersection` instead of `extract` we deal with 0--360
    longitudes automagically.

    Examples
    --------
    >>> import iris
    >>> url = "http://oos.soest.hawaii.edu/thredds/dodsC/pacioos/hycom/global"
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> bbox = [272.6, 24.25, 285.3, 36.70]
    >>> new_cube = subset(cube, bbox)
    >>> cube.shape != new_cube.shape
    True

    """
    if (cube.coord(axis='X').ndim == 1 and cube.coord(axis='Y').ndim == 1):
        # Workaround `cube.intersection` hanging up on FVCOM models.
        title = cube.attributes.get('title', 'untitled')
        featureType = cube.attributes.get('featureType', None)
        if (('FVCOM' in title) or ('ESTOFS' in title) or
           featureType == 'timeSeries'):
            cube = bbox_extract_1Dcoords(cube, bbox)
        else:
            cube = cube.intersection(longitude=(bbox[0], bbox[2]),
                                     latitude=(bbox[1], bbox[3]))
    elif (cube.coord(axis='X').ndim == 2 and
          cube.coord(axis='Y').ndim == 2):
        cube = bbox_extract_2Dcoords(cube, bbox)
    else:
        raise CoordinateMultiDimError(
            f'Cannot deal with '
            f'X:{cube.coord(axis="X").ndim} and '
            f'Y:{cube.coord(axis="y").ndim} dimensions.'
        )
    return cube


def _filter_none(lista):
    return [x for x in lista if x is not None]


def _in_list(cube, name_list):
    return cube.standard_name in name_list


def quick_load_cubes(url, name_list, callback=None, strict=False):
    """
    Return all cubes found using a `name_list` of standard_names.  The cubes
    found can be transformed via a `callback` function.
    If `strict` is set to True the function will return only one cube is
    possible, otherwise an exception will be raise.

    TODO: Create a criteria to choose a sensor.
    buoy = "http://129.252.139.124/thredds/dodsC/fldep.stlucieinlet..nc"
    buoy = "http://129.252.139.124/thredds/dodsC/lbhmc.cherrygrove.pier.nc"

    Examples
    --------
    >>> import iris
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> name_list = ['sea_water_potential_temperature']
    >>> cubes = quick_load_cubes(url, name_list)
    >>> cube = quick_load_cubes(url, name_list, strict=True)
    >>> isinstance(cubes, list)
    True
    >>> isinstance(cube, iris.cube.Cube)
    True

    """
    cubes = iris.load_raw(url, callback=callback)
    cubes = CubeList([cube for cube in cubes if _in_list(cube, name_list)])
    cubes = _filter_none(cubes)
    if not cubes:
        raise ValueError(f'Cannot find {name_list} in {url}.')
    if strict:
        if len(cubes) == 1:
            return cubes[0]
        else:
            ValueError(f'> 1 cube found! Expected just one.\n {repr(cubes)}')
    return cubes


def proc_cube(cube, bbox=None, time=None, constraint=None, units=None):
    """
    Constraining by `bbox`, `time`, and iris `constraint` object.
    and the `units` can be converted.

    Examples
    --------
    >>> import pytz
    >>> import iris
    >>> from datetime import date, datetime, timedelta
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> today = date.today().timetuple()
    >>> stop = datetime(today.tm_year, today.tm_mon, today.tm_mday, 12)
    >>> start = stop - timedelta(days=7)
    >>> bbox = [-87.40, 24.25, -74.70, 36.70]
    >>> name_list = ['sea_water_potential_temperature']
    >>> cube = quick_load_cubes(url, name_list, strict=True)
    >>> new_cube = proc_cube(cube, bbox=bbox, time=(start, stop))
    >>> cube.shape != new_cube.shape
    True

    """
    if constraint:
        cube = cube.extract(constraint)
        if not cube:
            raise ValueError(f'No cube using {constraint}')
    if bbox:
        cube = subset(cube, bbox)
        if not cube:
            raise ValueError(f'No cube using {bbox}')
    if time:
        if isinstance(time, datetime):
            start, stop = time, None
        elif isinstance(time, tuple):
            start, stop = time[0], time[1]
        else:
            raise ValueError(
                f'Time must be start or (start, stop).'
                '  Got {time}')
        cube = time_slice(cube, start, stop)
    if units:
        if cube.units != units:
            cube.convert_units(units)
    return cube


def add_mesh(cube, url):
    """
    Adds the unstructured mesh info the to cube.  Soon in an iris near you!

    """
    from pyugrid import UGrid
    ug = UGrid.from_ncfile(url)
    cube.mesh = ug
    cube.mesh_dimension = 1
    return cube


def ensure_timeseries(cube):
    """Ensure that the cube is CF-timeSeries compliant."""
    if not cube.coord('time').shape == cube.shape[0]:
        cube.transpose()
        [
            iris.util.demote_dim_coord_to_aux_coord(cube, dim)
            for dim in cube.dim_coords
            if 'time' not in dim.name()
        ]

    cube.attributes.update({'featureType': 'timeSeries'})
    cube.coord('station_code').attributes = {'cf_role': 'timeseries_id'}
    return cube


def add_station(cube, station):
    """Add a station Auxiliary Coordinate and its name."""
    kw = {'var_name': 'station', 'long_name': 'station_code'}
    coord = iris.coords.AuxCoord(station, **kw)
    cube.add_aux_coord(coord)
    return cube


def remove_ssh(cube):
    """
    Remove all `aux_coords` but time.  Should that has the same shape as
    the data.  NOTE: This also removes `aux_factories` to avoid update error
    when removing the coordinate.

    Examples
    --------
    >>> import iris
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> cube = get_surface(cube)
    >>> len(cube.coords())
    10
    >>> cube = remove_ssh(cube)
    >>> len(cube.coords())
    8

    """
    for factory in cube.aux_factories:
        cube.remove_aux_factory(factory)
    for coord in cube.aux_coords:
        if coord.shape == cube.shape:
            if 'time' not in coord.name():
                cube.remove_coord(coord.name())
    return cube


def cube2series(cube):
    """
    Take a cube data and metadata and return a rich `pandas.Series`.

    """
    time = cube.coord('time')
    index = time.units.num2date(time.points)
    data = cube.data.squeeze()
    series = pd.Series(data=data, index=index)
    series._metadata = {
        'station': cube.coord('station').points[0],
        'station_name': cube.coord('station_name').points[0],
        'station_code': cube.coord('station_code').points[0],
        'sensor': cube.coord('sensor').points[0],
        'lon': cube.coord('lon').points[0],
        'lat': cube.coord('lat').points[0],
        'depth': cube.coord('depth').points[0],
        'standard_name': cube.standard_name,
        'units': cube.units,
        }
    return series


def series2cube(series, attr=None):
    """
    Convert a metadata rich `pandas.Series` to a CF compliant timeSeries.
    http://cfconventions.org/Data/cf-convetions/cf-conventions-1.6/build/cf-conventions.html#idp5577536

    The metadata expected are: standard_name, units, station_code,
    station_name, station, sensor, lon, lat, depth and time.
    The last two are added as cube coords while the rest are added as either
    auxiliary coords or cube metadata.

    """

    data = np.ma.masked_invalid(series.values)
    cube = iris.cube.Cube(
        data=np.atleast_2d(data),
        standard_name=series._metadata['standard_name'],
        units=series._metadata['units'])
    if attr:
        cube.attributes.update(attr)

    _add_iris_coord(cube,
                    name='time',
                    points=series.index,
                    dim=1,
                    units=cf_units.Unit('hours since epoch',
                                        calendar=cf_units.CALENDAR_GREGORIAN))

    _add_iris_coord(cube,
                    name='depth',
                    points=np.float_(series._metadata['depth']),
                    dim=0,
                    units=cf_units.Unit('m'))

    _add_iris_coord(cube,
                    name='station_code',
                    points=str(series._metadata['station_code']),
                    dim=0, aux=True)

    _add_iris_coord(cube,
                    name='station',
                    points=str(series._metadata['station']),
                    dim=0, aux=True)

    _add_iris_coord(cube,
                    name='sensor',
                    points=str(series._metadata['sensor']),
                    dim=0, aux=True)

    _add_iris_coord(cube,
                    name='station_name',
                    points=str(series._metadata['station_name']),
                    dim=0, aux=True)

    _add_iris_coord(cube,
                    name='lon',
                    points=np.float_(series._metadata['lon']),
                    units=cf_units.Unit('degrees'),
                    dim=0, aux=True)

    _add_iris_coord(cube,
                    name='lat',
                    points=np.float_(series._metadata['lat']),
                    units=cf_units.Unit('degrees'),
                    dim=0, aux=True)
    return cube


def _add_iris_coord(cube, name, points, dim, units=None, aux=False):
    """
    Helper function to add a coordinate to an existing cube.

    """
    # Convert pandas datetime objects to python datetime obejcts.
    if isinstance(points, DatetimeIndex):
        points = points.to_pydatetime()

    # Convert datetime objects to Iris' current datetime representation.
    if isinstance(points, np.ndarray) and points.dtype == object:
        dt_types = (datetime, cftime.datetime)
        if all((isinstance(i, dt_types) for i in points)):
            points = units.date2num(points)

    points = np.array(points)
    if np.issubdtype(points.dtype, np.number) and not aux:
        coord = DimCoord(points, units=units)
        coord.rename(name)
        cube.add_dim_coord(coord, dim)
    else:
        coord = AuxCoord(points, units=units)
        coord.rename(name)
        cube.add_aux_coord(coord, dim)


def make_tree(cube):
    """
    Return a scipy KDTree object to search a cube.
    NOTE: iris does have its own implementation for searching with KDTrees, but
    it does not work for 2D coords like this one.

    Examples
    --------
    >>> import iris
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')
    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> cube = get_surface(cube)
    >>> tree, lon, lat = make_tree(cube)
    >>> isinstance(tree, KDTree)
    True

    """
    lon = cube.coord(axis='X').points
    lat = cube.coord(axis='Y').points
    # Structured models with 1D lon, lat.
    if (lon.ndim == 1) and (lat.ndim == 1) and (cube.ndim == 3):
        lon, lat = np.meshgrid(lon, lat)
    # Unstructured are already paired!
    tree = KDTree(list(zip(lon.ravel(), lat.ravel())))
    return tree, lon, lat


def is_water(cube, min_var=0.01):
    """
    Use only data where the standard deviation of the time cube exceeds
    0.01 m (1 cm) this eliminates flat line model time cube that come from
    land points that should have had missing values.
    (Accounts for wet-and-dry models.)

    """
    arr = np.ma.masked_invalid(cube.data).filled(fill_value=0)
    if arr.std() <= min_var:
        return False
    return True


def get_nearest_series(cube, tree, xi, yi, k=10, max_dist=0.04):
    """
    Find `k` nearest model data points from an iris `cube` at station
    lon: `xi`, lat: `yi` up to `max_dist` in degrees.  Must provide a Scipy's
    KDTree `tree`.

    Examples
    --------
    >>> import iris
    >>> from scipy.spatial import cKDTree as KDTree
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/'
    ...        'fmrc/us_east/US_East_Forecast_Model_Run_Collection_best.ncd')

    >>> cube = iris.load_cube(url, 'sea_water_potential_temperature')
    >>> cube = get_surface(cube)
    >>> tree, lon, lat = make_tree(cube)
    >>> series, dist, idx = get_nearest_series(cube, tree,
    ...                                        lon[0, 10], lat[0, 10],
    ...                                        k=10, max_dist=0.04)
    >>> idx == (0, 10)
    True
    >>> is_water(series, min_var=0.01)
    False
    >>> series, dist, idx = get_nearest_series(cube, tree,
    ...                                        -75.7135500943, 36.1640944084,
    ...                                        k=10, max_dist=0.04)
    >>> is_water(series, min_var=0.01)
    True

    """
    distances, indices = tree.query(np.array([xi, yi]).T, k=k)
    if indices.size == 0:
        raise ValueError('No data found.')
    # Get data up to specified distance.
    mask = distances <= max_dist
    distances, indices = distances[mask], indices[mask]
    if distances.size == 0:
        raise ValueError(f'No data near ({xi}, {yi}) max_dist={max_dist}.')
    # Unstructured model.
    if (cube.coord(axis='X').ndim == 1) and (cube.ndim == 2):
        i = j = indices
        unstructured = True
    # Structured model.
    else:
        unstructured = False
        if cube.coord(axis='X').ndim == 2:  # CoordinateMultiDim
            i, j = np.unravel_index(indices, cube.coord(axis='X').shape)
        else:
            shape = (cube.coord(axis='Y').shape[0],
                     cube.coord(axis='X').shape[0])
            i, j = np.unravel_index(indices, shape)
    series, dist, idx = None, None, None
    IJs = list(zip(i, j))
    for dist, idx in zip(distances, IJs):
        idx = tuple((int(kk) for kk in idx))
        if unstructured:  # NOTE: This would be so elegant in py3k!
            idx = (idx[0],)
        # This weird syntax allow for idx to be len 1 or 2.
        series = cube[(slice(None),)+idx]
    return series, dist, idx


def get_nearest_water(cube, tree, xi, yi, k=10, max_dist=0.04, min_var=0.01):
    """
    Legacy function.  Use `get_nearest_series`+`is_water` instead!

    """
    distances, indices = tree.query(np.array([xi, yi]).T, k=k)
    if indices.size == 0:
        raise ValueError('No data found.')
    # Get data up to specified distance.
    mask = distances <= max_dist
    distances, indices = distances[mask], indices[mask]
    if distances.size == 0:
        raise ValueError(f'No data near ({xi}, {yi}) max_dist={max_dist}.')
    # Unstructured model.
    if (cube.coord(axis='X').ndim == 1) and (cube.ndim == 2):
        i = j = indices
        unstructured = True
    # Structured model.
    else:
        unstructured = False
        if cube.coord(axis='X').ndim == 2:  # CoordinateMultiDim
            i, j = np.unravel_index(indices, cube.coord(axis='X').shape)
        else:
            shape = (cube.coord(axis='Y').shape[0],
                     cube.coord(axis='X').shape[0])
            i, j = np.unravel_index(indices, shape)
    IJs = list(zip(i, j))
    for dist, idx in zip(distances, IJs):
        idx = tuple((int(kk) for kk in idx))
        if unstructured:  # NOTE: This would be so elegant in py3k!
            idx = (idx[0],)
        # This weird syntax allow for idx to be len 1 or 2.
        series = cube[(slice(None),)+idx]
        if is_water(series, min_var=0.01):
            break
        else:
            series = None
            continue
    return series, dist, idx



import matplotlib.pyplot as plt

import numpy as np


__all__ = ['TaylorDiagram']


class TaylorDiagram(object):
    """
    Taylor diagram: plot model standard deviation and correlation
    to reference (data) sample in a single-quadrant polar plot, with
    r=stddev and theta=arccos(correlation).

    Taylor diagram (Taylor, 2001) test implementation.

    http://www-pcmdi.llnl.gov/about/staff/Taylor/CV/Taylor_diagram_primer.htm
    https://gist.github.com/ycopin/3342888

    """

    def __init__(self, refstd, fig=None, rect=111, label='_'):
        """
        Set up Taylor diagram axes, i.e. single quadrant polar
        plot, using mpl_toolkits.axisartist.floating_axes. refstd is
        the reference standard deviation to be compared to.

        """

        from matplotlib.projections import PolarAxes
        from mpl_toolkits.axisartist import floating_axes
        from mpl_toolkits.axisartist import grid_finder

        self.refstd = refstd  # Reference standard deviation.

        tr = PolarAxes.PolarTransform()

        # Correlation labels.
        rlocs = np.concatenate((np.arange(10)/10., [0.95, 0.99]))
        tlocs = np.arccos(rlocs)  # Conversion to polar angles.
        gl1 = grid_finder.FixedLocator(tlocs)  # Positions.
        dict_formatter = {
            list(zip(tlocs, map(str, rlocs)))
            }
        tf1 = grid_finder.DictFormatter(dict_formatter)

        # Standard deviation axis extent.
        self.smin = 0
        self.smax = 1.5*self.refstd

        extremes = (0,
                    np.pi/2,  # 1st quadrant.
                    self.smin,
                    self.smax)
        ghelper = floating_axes.GridHelperCurveLinear(tr,
                                                      extremes=extremes,
                                                      grid_locator1=gl1,
                                                      tick_formatter1=tf1)

        if fig is None:
            fig = plt.figure()

        ax = floating_axes.FloatingSubplot(fig, rect, grid_helper=ghelper)
        fig.add_subplot(ax)

        # Adjust axes.
        ax.axis['top'].set_axis_direction('bottom')  # Angle axis.
        ax.axis['top'].toggle(ticklabels=True, label=True)
        ax.axis['top'].major_ticklabels.set_axis_direction('top')
        ax.axis['top'].label.set_axis_direction('top')
        ax.axis['top'].label.set_text('Correlation')

        ax.axis['left'].set_axis_direction('bottom')  # X axis.
        ax.axis['left'].label.set_text('Standard deviation')

        ax.axis['right'].set_axis_direction('top')  # Y axis.
        ax.axis['right'].toggle(ticklabels=True)
        ax.axis['right'].major_ticklabels.set_axis_direction('left')

        ax.axis['bottom'].set_visible(False)  # Useless.

        # Contours along standard deviations.
        ax.grid(False)

        self._ax = ax  # Graphical axes.
        self.ax = ax.get_aux_axes(tr)  # Polar coordinates.

        # Add reference point and stddev contour.
        l, = self.ax.plot([0], self.refstd, 'ko',
                          ls='', ms=10, label=label)
        t = np.linspace(0, np.pi/2)
        r = np.zeros_like(t) + self.refstd
        self.ax.plot(t, r, 'k--', label='_')

        # Collect sample points for latter use (e.g. legend).
        self.samplePoints = [l]

    def add_sample(self, stddev, corrcoef, *args, **kwargs):
        """
        Add sample (stddev, corrcoeff) to the Taylor diagram. args
        and kwargs are directly propagated to the Figure.plot
        command.

        """
        l, = self.ax.plot(np.arccos(corrcoef), stddev,
                          *args, **kwargs)  # (theta, radius).
        self.samplePoints.append(l)
        return l

    def add_contours(self, levels=5, **kwargs):
        """
        Add constant centered RMS difference contours.

        """
        rs, ts = np.meshgrid(np.linspace(self.smin, self.smax),
                             np.linspace(0, np.pi/2))
        # Compute centered RMS difference.
        rms = np.sqrt(self.refstd**2 + rs**2 - 2*self.refstd*rs*np.cos(ts))
        contours = self.ax.contour(ts, rs, rms, levels, **kwargs)
        return contours


if __name__ == '__main__':
    # Reference dataset.
    x = np.linspace(0, 4*np.pi, 100)
    data = np.sin(x)
    refstd = data.std(ddof=1)  # Reference standard deviation.

    # Models.
    m1 = data + 0.2*np.random.randn(len(x))  # Model 1.
    m2 = 0.8*data + 0.1*np.random.randn(len(x))  # Model 2.
    m3 = np.sin(x-np.pi/10)  # Model 3.

    # Compute stddev and correlation coefficient of models.
    samples = np.array([[m.std(ddof=1), np.corrcoef(data, m)[0, 1]]
                        for m in (m1, m2, m3)])

    fig = plt.figure(figsize=(10, 4))

    ax1 = fig.add_subplot(1, 2, 1, xlabel='X', ylabel='Y')
    # Taylor diagram.
    dia = TaylorDiagram(refstd, fig=fig, rect=122, label='Reference')

    colors = plt.matplotlib.cm.jet(np.linspace(0, 1, len(samples)))

    ax1.plot(x, data, 'ko', label='Data')
    prop = {'size': 'small'}
    for k, m in enumerate([m1, m2, m3]):
        ax1.plot(x, m, c=colors[k], label='Model %d' % (k+1))
    ax1.legend(numpoints=1, prop=prop, loc='best')

    # Add samples to Taylor diagram.
    for k, (stddev, corrcoef) in enumerate(samples):
        dia.add_sample(
            stddev,
            corrcoef,
            marker='s',
            ls='',
            c=colors[k],
            label=f'Model {k+1}'
            )

    # Add RMS contours, and label them.
    contours = dia.add_contours(colors='0.5')
    plt.clabel(contours, inline=1, fontsize=10)

    # Add a figure legend.
    fig.legend(dia.samplePoints,
               [p.get_label() for p in dia.samplePoints],
               numpoints=1, prop=prop, loc='upper right')
    plt.show()



import fnmatch
import os
from contextlib import contextmanager
from glob import glob
from io import BytesIO
from urllib.parse import urlparse
from urllib.request import urlopen

import iris
from iris.pandas import as_data_frame

from lxml import etree

import numpy as np

from owslib import fes
from owslib.ows import ExceptionReport

import pandas as pd

import requests

from retrying import retry

import timeout_decorator

from .tardis import cube2series

"""
Collection of functions used in the IOOS system-test exercise.

"""


def parse_config(config_file):
    """Parse the yaml file with the configuration for each run."""
    from datetime import date, datetime, timedelta

    import yaml
    import pytz
    import cf_units

    with open(config_file, 'r') as f:
        config = yaml.safe_load(f)

    # Dates are normalized to UTC.
    start = config['date']['start']
    stop = config['date']['stop']
    if isinstance(start, int) and isinstance(stop, int):
        start = datetime.combine(
            date.today() - timedelta(days=abs(start)), datetime.min.time()
            )
        stop = datetime.combine(
            date.today() + timedelta(days=abs(stop)), datetime.min.time()
            )
    elif isinstance(start, int) and isinstance(stop, datetime):
        start = stop - timedelta(days=abs(start))
    elif isinstance(start, datetime) and isinstance(stop, int):
        stop = start + timedelta(days=abs(stop))
    elif isinstance(start, datetime) and isinstance(stop, datetime):
        pass
    else:
        raise ValueError(
            f'Expect dates (YYYY-MM-DD hh:mm:ss) or days offest (int).'
            '\nGot start={start} and stop={stop}.'
            )
    config['date']['start'] = start.replace(tzinfo=pytz.utc)
    config['date']['stop'] = stop.replace(tzinfo=pytz.utc)

    # Units.
    config['units'] = cf_units.Unit(config['units'])

    return config


def fes_date_filter(start, stop, constraint='overlaps'):
    """
    Take datetime-like objects and returns a fes filter for date range
    (begin and end inclusive).
    NOTE: Truncates the minutes!!!

    Examples
    --------
    >>> import pytz
    >>> from datetime import datetime, timedelta
    >>> stop = datetime(2010, 1, 1, 12, 30, 59).replace(tzinfo=pytz.utc)
    >>> start = stop - timedelta(days=7)
    >>> begin, end = fes_date_filter(start, stop, constraint='overlaps')
    >>> begin.literal, end.literal
    ('2010-01-01 12:00', '2009-12-25 12:00')
    >>> begin.propertyoperator, end.propertyoperator
    ('ogc:PropertyIsLessThanOrEqualTo', 'ogc:PropertyIsGreaterThanOrEqualTo')
    >>> begin, end = fes_date_filter(start, stop, constraint='within')
    >>> begin.literal, end.literal
    ('2009-12-25 12:00', '2010-01-01 12:00')
    >>> begin.propertyoperator, end.propertyoperator
    ('ogc:PropertyIsGreaterThanOrEqualTo', 'ogc:PropertyIsLessThanOrEqualTo')

    """
    start = start.strftime('%Y-%m-%d %H:00')
    stop = stop.strftime('%Y-%m-%d %H:00')
    if constraint == 'overlaps':
        propertyname = 'apiso:TempExtent_begin'
        begin = fes.PropertyIsLessThanOrEqualTo(propertyname=propertyname,
                                                literal=stop)
        propertyname = 'apiso:TempExtent_end'
        end = fes.PropertyIsGreaterThanOrEqualTo(propertyname=propertyname,
                                                 literal=start)
    elif constraint == 'within':
        propertyname = 'apiso:TempExtent_begin'
        begin = fes.PropertyIsGreaterThanOrEqualTo(propertyname=propertyname,
                                                   literal=start)
        propertyname = 'apiso:TempExtent_end'
        end = fes.PropertyIsLessThanOrEqualTo(propertyname=propertyname,
                                              literal=stop)
    else:
        raise NameError('Unrecognized constraint {}'.format(constraint))
    return begin, end


def get_csw_records(csw, filter_list, pagesize=10, maxrecords=1000, **kwargs):
    """
    Iterate `maxrecords`/`pagesize` times until the requested value in
    `maxrecords` is reached.

    `kwargs` can by any `getrecords2` key word argument, like esn='full'.

    """
    from owslib.fes import SortBy, SortProperty

    # Iterate over sorted results.
    sortby = SortBy([SortProperty('dc:title', 'ASC')])
    csw_records = {}
    startposition = 0

    nextrecord = getattr(csw, 'results', 1)
    while nextrecord != 0:
        csw.getrecords2(
            constraints=filter_list,
            startposition=startposition,
            maxrecords=pagesize,
            sortby=sortby,
            **kwargs
            )
        csw_records.update(csw.records)
        if csw.results['nextrecord'] == 0:
            break
        startposition += pagesize + 1  # Last one is included.
        if startposition >= maxrecords:
            break
    csw.records.update(csw_records)
    return csw


def _parse_reference(ref, identifier):
    """
    First try to sniff the scheme from the URL in the `ref` dict with geolinks,
    if that fails get the `scheme` field directly.

    For all possible identifiers see:
    https://github.com/OSGeo/Cat-Interop/blob/master/LinkPropertyLookupTable.csv

    """
    from geolinks import sniff_link

    url = None
    scheme = sniff_link(ref['url'])
    if not scheme:
        scheme = ref['scheme']
    if identifier.endswith(':'):
        cond = identifier in scheme
    else:
        cond = identifier == scheme
    if cond:
        url = ref['url']
    return url


def service_urls(records, identifier='OGC:SOS'):
    """
    Extract service ULRs from csw records using geolink identifiers
    (OPeNDAP:OPeNDAP, ERDDAP:griddap, ERDDAP:tabledap, OGC:SOS, etc).


    For all possible identifiers see:
    https://github.com/OSGeo/Cat-Interop/blob/master/LinkPropertyLookupTable.csv

    If is possible to truncate ambiguous identifiers at the `:`,
    and return everything that startswith that identifier.

    """
    urls = []
    for key, rec in records.items():
        for ref in rec.references:
            url = _parse_reference(ref, identifier)
            if url:
                urls.append(url)
    return sorted(set(urls))


def sos_request(url='opendap.co-ops.nos.noaa.gov/ioos-dif-sos/SOS', **kw):
    """
    Examples
    --------
    >>> try:
    ...     from urlparse import urlparse
    ... except ImportError:
    ...     from urllib.parse import urlparse
    >>> from datetime import date, datetime, timedelta
    >>> today = date.today().strftime("%Y-%m-%d")
    >>> start = datetime.strptime(today, "%Y-%m-%d") - timedelta(7)
    >>> bbox = [-87.40, 24.25, -74.70, 36.70]
    >>> sos_name = 'water_surface_height_above_reference_datum'
    >>> offering='urn:ioos:network:NOAA.NOS.CO-OPS:WaterLevelActive'
    >>> params = {'observedProperty': sos_name,
    ...           'eventTime': start.strftime('%Y-%m-%dT%H:%M:%SZ'),
    ...           'featureOfInterest': 'BBOX:{0},{1},{2},{3}'.format(*bbox),
    ...           'offering': offering}
    >>> uri = 'http://opendap.co-ops.nos.noaa.gov/ioos-dif-sos/SOS'
    >>> url = sos_request(uri, **params)
    >>> bool(urlparse(url).scheme)
    True

    """
    url = parse_url(url)
    offering = 'urn:ioos:network:NOAA.NOS.CO-OPS:CurrentsActive'
    params = {
        'service': 'SOS',
        'request': 'GetObservation',
        'version': '1.0.0',
        'offering': offering,
        'responseFormat': 'text/csv'
    }
    params.update(kw)
    r = requests.get(url, params=params)
    r.raise_for_status()
    content = r.headers['Content-Type']
    if 'excel' in content or 'csv' in content:
        return r.url
    else:
        raise TypeError('Bad url {}'.format(r.url))


def _get_value(sensor, name='longName'):
    value = None
    sml = sensor.get(name, None)
    if sml:
        value = sml.value
    return value


def get_coops_metadata(station):
    """
    Get longName and sensorName for specific station from COOPS SOS using
    DescribeSensor and owslib.swe.sensor.sml.SensorML.

    Examples
    --------
    >>> long_name, station_id = get_coops_metadata(8651370)
    >>> long_name
    'Duck, NC'
    >>> station_id
    'urn:ioos:station:NOAA.NOS.CO-OPS:8651370'

    """
    from owslib.swe.sensor.sml import SensorML
    url = ('opendap.co-ops.nos.noaa.gov/ioos-dif-sos/SOS?'
           'service=SOS&'
           'request=DescribeSensor&version=1.0.0&'
           'outputFormat=text/xml;'
           'subtype="sensorML/1.0.1/profiles/ioos_sos/1.0"&'
           'procedure=urn:ioos:station:NOAA.NOS.CO-OPS:%s') % station
    url = parse_url(url)
    xml = etree.parse(urlopen(url))
    root = SensorML(xml)
    if not root.members or len(root.members) > 1:
        raise ValueError(f'Expected 1 member, got {len(root.members)}')
    system = root.members[0]

    # NOTE: Some metadata of interest.
    # system.description
    # short_name = _get_value(system.identifiers, name='shortName')
    # [c.values() for c in system.components]

    long_name = _get_value(system.identifiers, name='longName')
    long_name = long_name.split('station, ')[-1].strip()
    station_id = _get_value(system.identifiers, name='stationID')

    return long_name, station_id


def ndbc2df(collector, ndbc_id):
    """
    Ugly hack because `collector.raw(responseFormat="text/csv")`
    Usually times out.

    """
    from netCDF4 import MFDataset, date2index, num2date
    # FIXME: Only sea_water_temperature for now.
    if len(collector.variables) > 1:
        raise ValueError(
            f'Expected only 1 variable to download, got {collector.variables}'
        )
    if collector.variables[0] == 'sea_water_temperature':
        columns = 'sea_water_temperature (C)'
        ncvar = 'sea_surface_temperature'
        data_type = 'stdmet'
        # adcp, adcp2, cwind, dart, mmbcur, ocean, oceansites, pwind,
        # swden, tao-ctd, wlevel, z-hycom
    else:
        raise ValueError(f'Do not know how to download {collector.variables}')

    uri = 'http://dods.ndbc.noaa.gov/thredds/dodsC/data/{}'.format(data_type)
    url = ('%s/%s/' % (uri, ndbc_id))
    urls = url_lister(url)

    filetype = '*.nc'
    file_list = [filename for filename in fnmatch.filter(urls, filetype)]
    files = [fname.split('/')[-1] for fname in file_list]
    urls = ['%s/%s/%s' % (uri, ndbc_id, fname) for fname in files]

    if not urls:
        raise Exception(f'Cannot find data at {url}')
    nc = MFDataset(urls)

    kw = {'calendar': 'gregorian', 'select': 'nearest'}
    time_dim = nc.variables['time']
    time = num2date(time_dim[:], units=time_dim.units,
                    calendar=kw['calendar'])

    idx_start = date2index(collector.start_time.replace(tzinfo=None),
                           time_dim, **kw)
    idx_stop = date2index(collector.end_time.replace(tzinfo=None),
                          time_dim, **kw)
    if idx_start == idx_stop:
        raise Exception(
            'No data within time range '
            ' {collector.start_time} and {collector.end_time}'
        )
    data = nc.variables[ncvar][idx_start:idx_stop, ...].squeeze()

    time_dim = nc.variables['time']
    time = time[idx_start:idx_stop].squeeze()
    df = pd.DataFrame(data=data, index=time, columns=[columns])
    df.index.name = 'date_time'
    return df


def pyoos2df(collector, station_id, df_name=None):
    """
    Request CSV response from SOS and convert to Pandas dataframe.

    """
    collector.features = [station_id]
    try:
        response = collector.raw(responseFormat='text/csv')
        kw = {'parse_dates': True, 'index_col': 'date_time'}
        df = pd.read_csv(BytesIO(response), **kw)
    except requests.exceptions.ReadTimeout:
        df = ndbc2df(collector, station_id)
    # FIXME: Workaround to get only 1 sensor.
    df = df.reset_index()
    kw = {'subset': 'date_time', 'keep': 'last'}
    df = df.drop_duplicates(**kw).set_index('date_time')
    if df_name:
        df.name = df_name
    return df


@retry(stop_max_attempt_number=5, wait_fixed=3000)
def collector2table(collector, config, col='sea_water_temperature (C)'):
    """
    collector2table returns the station stable as a DataFrame.
    columns are station, sensor, lon, lat, and the index is the station
    number.

    Alternative to `ndbc2df` and `pyoos2df`.

    """
    import copy

    c = copy.copy(collector)
    c.features = None
    try:
        response = c.raw(responseFormat='text/csv')
    except ExceptionReport:
        try:
            end = c.end_time
            response = c.filter(end=c.start_time).raw(responseFormat='text/csv')  # noqa
            c.filter(end=end)
        except ExceptionReport:
            # No data available in collection, so return an empty list.
            return []
    df = pd.read_csv(BytesIO(response), parse_dates=True)
    g = df.groupby('station_id')
    df = {}
    for station in g.groups.keys():
        df.update({station: g.get_group(station).iloc[0]})
    df = pd.DataFrame.from_dict(df).T

    station_dict = {}
    for offering in c.server.offerings:
        station_dict.update({offering.name: offering.description})

    names = []
    for sta in df.index:
        names.append(station_dict.get(sta, sta))

    df['name'] = names

    observations = []
    for k, row in df.iterrows():
        station_id = row['station_id'].split(':')[-1]
        c.features = [station_id]
        response = c.raw(responseFormat='text/csv')
        kw = {'parse_dates': True, 'index_col': 'date_time'}
        data = pd.read_csv(BytesIO(response), **kw).reset_index()
        data = data.drop_duplicates(subset='date_time').set_index('date_time')
        series = data[col]
        series._metadata = {
            'station': row.get('station_id'),
            'station_name': row.get('name'),
            'station_code': str(row.get('station_id').split(':')[-1]),
            'sensor': row.get('sensor_id'),
            'lon': row.get('longitude (degree)'),
            'lat': row.get('latitude (degree)'),
            'depth': row.get('depth (m)'),
            'standard_name': config['sos_name'],
            'units': config['units'],
        }

        observations.append(series)
    return observations


def _extract_columns(name, cube):
    """
    Workaround to extract data from a cube and create a dataframe
    following SOS boilerplate.

    """
    try:
        from HTMLParser import HTMLParser
    except ImportError:
        from html.parser import HTMLParser
    station = cube.attributes.get('abstract', None)
    if not station:
        station = name.replace('.', '_')

    parser = HTMLParser()
    station = parser.unescape(station)

    sensor = 'NA'
    lon = cube.coord(axis='X').points[0]
    lat = cube.coord(axis='Y').points[0]
    time = cube.coord(axis='T')
    time = time.units.num2date(cube.coord(axis='T').points)[0]
    date_time = time.strftime('%Y-%M-%dT%H:%M:%SZ')
    data = cube.data.mean()
    return station, sensor, lat, lon, date_time, data


def secoora2df(buoys, varname):
    secoora_obs = {}
    for station, cube in buoys.items():
        secoora_obs.update({station: _extract_columns(station, cube)})

    df = pd.DataFrame.from_dict(secoora_obs, orient='index')
    df.reset_index(inplace=True)
    columns = {'index': 'station',
               0: 'name',
               1: 'sensor',
               2: 'lat',
               3: 'lon',
               4: 'date_time',
               5: varname}

    df.rename(columns=columns, inplace=True)
    df.set_index('name', inplace=True)
    return df


def _guess_name(model_full_name):
    """
    Examples
    --------
    >>> some_names = ['USF FVCOM - Nowcast Aggregation',
    ...               'ROMS/TOMS 3.0 - New Floria Shelf Application',
    ...               'COAWST Forecast System : USGS : US East Coast and Gulf'
    ...               'of Mexico (Experimental)',
    ...               'HYbrid Coordinate Ocean Model (HYCOM): Global',
    ...               'ROMS ESPRESSO Real-Time Operational IS4DVAR Forecast'
    ...               'System Version 2 (NEW) 2013-present FMRC History'
    ...               '(Best)']
    >>> [_guess_name(model_full_name) for model_full_name in some_names]
    ['USF_FVCOM', 'ROMS/TOMS', 'COAWST_USGS', 'HYCOM', 'ROMS_ESPRESSO']

    """
    words = []
    for word in model_full_name.split():
        if word.isupper():
            words.append(_remove_parenthesis(word))
    mod_name = ' '.join(words)
    if not mod_name:
        mod_name = ''.join([c for c in model_full_name.split('(')[0]
                            if c.isupper()])
    if len(mod_name.split()) > 1:
        mod_name = '_'.join(mod_name.split()[:2])
    return mod_name


def _remove_parenthesis(word):
    """
    Examples
    --------
    >>> _remove_parenthesis('(ROMS)')
    'ROMS'

    """
    try:
        return word[word.index('(') + 1:word.rindex(')')]
    except ValueError:
        return word


def _sanitize(name):
    """
    Examples
    --------
    >>> _sanitize('ROMS/TOMS')
    'ROMS_TOMS'
    >>> _sanitize('USEAST model')
    'USEAST_model'
    >>> _sanitize('GG1SST, SST')
    'GG1SST_SST'

    """
    name = name.replace(', ', '_')
    name = name.replace('/', '_')
    name = name.replace(' ', '_')
    name = name.replace(',', '_')
    return name


def get_model_name(url):
    """
    Return a model short name based on its endpoint.

    Examples
    --------
    >>> url = ('http://omgsrv1.meas.ncsu.edu:8080/thredds/dodsC/fmrc/sabgom/'
    ...        'SABGOM_Forecast_Model_Run_Collection_best.ncd')
    >>> get_model_name(url)
    'fmrc-SABGOM_Forecast_Model_Run_Collection_best'

    """
    names = url.split('dodsC/')[-1].split('/')
    names, fname = names[:-1], names[-1]
    names = [name for name in names if name.lower() not in fname.lower()]

    if fname.endswith('.ncd') or fname.endswith('.nc'):
        fname = fname.rstrip('.ncd').rstrip('.nc')

    if names:
        mod_name = '{}-{}'.format('_'.join(names), fname)
    else:
        mod_name = '{}'.format(fname)
    return mod_name


@timeout_decorator.timeout(10, use_signals=False)
def is_station(url):
    from netCDF4 import Dataset
    with Dataset(url) as nc:
        station = False
        if hasattr(nc, 'cdm_data_type'):
            if nc.cdm_data_type.lower() == 'station':
                station = True
    return station


def nc2df(fname, columns_name='station_code'):
    """
    Load a netCDF timeSeries file as a dataframe.

    """
    cube = iris.load_cube(fname)
    for coord in cube.coords(dimensions=[0]):
        name = coord.name()
        if name != 'time':
            cube.remove_coord(name)
    for coord in cube.coords(dimensions=[1]):
        name = coord.name()
        if name != columns_name:
            cube.remove_coord(name)
    df = as_data_frame(cube)
    if cube.ndim == 1:  # Horrible work around iris.
        station = cube.coord(columns_name).points[0]
        df.columns = [station]
    return df


def stations_keys(config, key='station_name'):
    save_dir = os.path.join(os.path.abspath(config['run_name']))
    fname = os.path.join(save_dir, '{}.nc'.format('OBS_DATA'))
    cubes = iris.load_raw(fname)
    observations = [cube2series(cube) for cube in cubes]
    return {obs._metadata['station_code']: obs._metadata[key] for
            obs in observations}


def load_ncs(config):
    save_dir = os.path.join(os.path.abspath(config['run_name']))
    fname = '{}.nc'.format
    fname = os.path.join(save_dir, fname('OBS_DATA'))

    cubes = iris.load_raw(fname)
    data = [cube2series(cube) for cube in cubes]
    index = pd.date_range(start=config['date']['start'].replace(tzinfo=None),
                          end=config['date']['stop'].replace(tzinfo=None),
                          freq='1H')
    # Preserve metadata with `reindex`.
    observations = []
    for series in data:
        _metadata = series._metadata
        obs = series.reindex(index=index, limit=1, method='nearest')
        obs._metadata = _metadata
        observations.append(obs)

    for obs in observations:
        obs.name = obs._metadata.get('station_code')
    ALL_OBS_DATA = pd.DataFrame(observations).T

    dfs = {'OBS_DATA': ALL_OBS_DATA}
    for fname in glob(os.path.join(config['run_name'], '*.nc')):
        if 'OBS_DATA' in fname:
            continue
        else:
            model = os.path.splitext(os.path.split(fname)[-1])[0].split('-')[-1]  # noqa
            df = nc2df(fname, columns_name='station_code')
            # FIXME: Horrible work around duplicate times.
            if len(df.index.values) != len(np.unique(df.index.values)):
                kw = {'subset': 'index', 'take_last': True}
                df = df.reset_index().drop_duplicates(**kw).set_index('index')
            kw = {'method': 'time', 'limit': 2}
            df = df.reindex(index).interpolate(**kw).ix[index]
            dfs.update({model: df})
    kw = {'orient': 'items', 'intersect': False}
    dfs = pd.Panel.from_dict(dfs, **kw).swapaxes(0, 2)
    return dfs


# Web/Misc.
def url_lister(url):
    """
    Extract all href links from a given URL.

    """
    import lxml.html
    urls = []
    connection = urlopen(url)
    dom = lxml.html.fromstring(connection.read())
    for link in dom.xpath('//a/@href'):
        urls.append(link)
    return urls


def parse_url(url):
    """
    This will preserve any given scheme but will add http if none is
    provided.

    Examples
    --------
    >>> parse_url('www.google.com')
    'http://www.google.com'
    >>> parse_url('https://www.google.com')
    'https://www.google.com'

    """
    if not urlparse(url).scheme:
        url = f'http://{url}'
    return url


def to_html(df, css=None):
    """
    Return a pandas table HTML representation with the datagrid css.
    Examples
    --------
    >>> from IPython.display import HTML
    >>> from pandas import DataFrame
    >>> df = DataFrame(np.empty((5, 5)))
    >>> html = to_html(df)
    >>> isinstance(html, HTML)
    True

    """
    from IPython.display import HTML
    if css:
        style = f"""<style>{css}</style>"""
    else:
        style = ''
    table = {'style': style, 'table': df.to_html()}
    return HTML(f'{style}<div class="datagrid">{table}</div>')


def save_html(fname, HTML):
    with open(fname, 'w') as f:
        f.writelines(HTML.data)


@contextmanager
def time_limit(seconds=10):
    """
    Raise a TimeoutException after n `seconds`.

    """
    import signal

    def signal_handler(signum, frame):
        raise TimeoutException('Timed out!')
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


class TimeoutException(Exception):
    """
    Timeout Exception.

    Example
    -------
    >>> def long_function_call():
    ...     import time
    ...     sec = 0
    ...     while True:
    ...         sec += 1
    ...         time.sleep(1)
    >>> try:
    ...     with time_limit(3):
    ...         long_function_call()
    ... except TimeoutException as msg:
    ...     print('{!r}'.format(msg))
    TimeoutException('Timed out!',)
    """
    pass


def make_map(bbox, **kw):
    """
    Creates a folium map instance for SECOORA.

    Examples
    --------
    >>> from folium import Map
    >>> bbox = [-87.40, 24.25, -74.70, 36.70]
    >>> m = make_map(bbox)
    >>> isinstance(m, Map)
    True

    """
    import folium

    line = kw.pop('line', True)
    layers = kw.pop('layers', True)
    hf_radar = kw.pop('hf_radar', True)
    zoom_start = kw.pop('zoom_start', 5)

    lon, lat = np.array(bbox).reshape(2, 2).mean(axis=0)
    m = folium.Map(width='100%', height='100%',
                   location=[lat, lon], zoom_start=zoom_start)

    if hf_radar:
        url = 'http://hfrnet.ucsd.edu/thredds/wms/HFRNet/USEGC/6km/hourly/RTV'
        w = folium.WmsTileLayer(url,
                                name='HF Radar',
                                fmt='image/png',
                                layers='surface_sea_water_velocity',
                                attr='HFRNet',
                                overlay=True,
                                transparent=True)
        w.add_to(m)
    if layers:
        add = 'MapServer/tile/{z}/{y}/{x}'
        base = 'http://services.arcgisonline.com/arcgis/rest/services'
        ESRI = {
            'Imagery': 'World_Imagery/MapServer',
            'Ocean_Base': 'Ocean/World_Ocean_Base',
            'Topo_Map': 'World_Topo_Map/MapServer',
            'Street_Map': 'World_Street_Map/MapServer',
            'Physical_Map': 'World_Physical_Map/MapServer',
            'Terrain_Base': 'World_Terrain_Base/MapServer',
            'NatGeo_World_Map': 'NatGeo_World_Map/MapServer',
            'Shaded_Relief': 'World_Shaded_Relief/MapServer',
            'Ocean_Reference': 'Ocean/World_Ocean_Reference',
            'Navigation_Charts': 'Specialty/World_Navigation_Charts'
            }
        for name, url in ESRI.items():
            url = '{}/{}/{}'.format(base, url, add)

            w = folium.TileLayer(tiles=url,
                                 name=name,
                                 attr='ESRI',
                                 overlay=True)
            w.add_to(m)

    if line:  # Create the map and add the bounding box line.
        p = folium.PolyLine(get_coordinates(bbox),
                            color='#FF0000',
                            weight=2,
                            opacity=0.9,
                            latlon=True)
        p.add_to(m)

    folium.LayerControl().add_to(m)
    return m


def get_coordinates(bbox):
    """
    Create bounding box coordinates for the map.  It takes flat or
    nested list/numpy.array and returns 5 points that closes square
    around the borders.

    Examples
    --------
    >>> bbox = [-87.40, 24.25, -74.70, 36.70]
    >>> len(get_coordinates(bbox))
    5

    """
    bbox = np.asanyarray(bbox).ravel()
    if bbox.size == 4:
        bbox = bbox.reshape(2, 2)
        coordinates = []
        coordinates.append([bbox[0][1], bbox[0][0]])
        coordinates.append([bbox[0][1], bbox[1][0]])
        coordinates.append([bbox[1][1], bbox[1][0]])
        coordinates.append([bbox[1][1], bbox[0][0]])
        coordinates.append([bbox[0][1], bbox[0][0]])
    else:
        raise ValueError('Wrong number corners.'
                         '  Expected 4 got {}'.format(bbox.size))
    return coordinates



import numpy as np
import numpy.ma as ma

from pandas import DataFrame


def both_valid(x, y):
    """
    Returns a mask where both series are valid.

    Examples
    --------
    >>> import numpy as np
    >>> x = [np.NaN, 1, 2, 3, 4, 5]
    >>> y = [0, 1, np.NaN, 3, 4, 5]
    >>> both_valid(x, y)
    array([False,  True, False,  True,  True,  True])

    """
    mask_x = np.isnan(x)
    mask_y = np.isnan(y)
    return np.logical_and(~mask_x, ~mask_y)


def mean_bias(y_true, y_pred,
              sample_weight=None,
              multioutput='uniform_average'):
    """Modified from `mean_absolute_error` to preserve the bias sign."""
    from sklearn.metrics.regression import _check_reg_targets

    y_type, y_true, y_pred, multioutput = _check_reg_targets(
        y_true, y_pred, multioutput)
    output_errors = np.average(np.abs(y_pred - y_true),
                               weights=sample_weight, axis=0)
    output_errors *= np.sign(np.average(y_pred - y_true))
    if multioutput == 'raw_values':
        return output_errors
    elif multioutput == 'uniform_average':
        # Pass None as weights to np.average: uniform mean.
        multioutput = None
    return np.average(output_errors, weights=multioutput)


def mean_absolute_bias(obs, model):
    from sklearn.metrics import mean_absolute_error
    return mean_absolute_error(obs, model)


def median_absolute_bias(obs, model):
    from sklearn.metrics import median_absolute_error
    return median_absolute_error(obs, model)


def rmse(obs, model):
    """
    Compute root mean square between the observed data (`obs`) and the modeled
    data (`model`).
    >>> obs = [3, -0.5, 2, 7]
    >>> model = [2.5, 0.0, 2, 8]
    >>> rmse(obs, model)
    0.6123724356957945
    >>> obs = [[0.5, 1],[-1, 1],[7, -6]]
    >>> model = [[0, 2],[-1, 2],[8, -5]]
    >>> rmse(obs, model)
    0.8416254115301732

    """
    from sklearn.metrics import mean_squared_error
    return np.sqrt(mean_squared_error(obs, model))


def r2(x, y):
    from sklearn.metrics import r2_score
    return r2_score(x, y)


def apply_skill(dfs, function, remove_mean=True, filter_tides=False):
    skills = {}
    for station, df in dfs.iteritems():
        if filter_tides:
            df = df.apply(low_pass)
        skill = {}
        obs = df.pop('OBS_DATA')
        if obs.isnull().all():
            # No observations.
            skills.update({station: np.NaN})
            continue
        for model, y in df.iteritems():
            # No models.
            if y.isnull().all():
                skills.update({station: np.NaN})
                continue
            mask = both_valid(obs, y)
            x, y = obs[mask], y[mask]
            if remove_mean:
                x, y = x-x.mean(), y-y.mean()
            if x.size:
                ret = function(x, y)
            else:
                ret = np.NaN
            skill.update({model: ret})
        skills.update({station: skill})
    return DataFrame.from_dict(skills)


def low_pass(series, window_size=193, T=40, dt=360):
    """
    This function applies a lanczos filter on a pandas time-series and
    returns the low pass data series.

    series : Pandas series
    window_size : Size of the filter windows (default is 96+1+96).
    T : Period of the filter. (The default of 40 hours filter should
        get rid of all tides.)
    dt : time_delta in seconds.  Default is 360 (6 minutes).

    """
    from oceans import lanc
    T *= 60*60.  # To seconds.
    freq = dt/T

    mask = np.isnan(series)
    avg = series.mean()
    series = series - avg
    series.interpolate(inplace=True)

    wt = lanc(window_size, freq)
    low = np.convolve(wt, series, mode='same')
    low = ma.masked_array(low, mask)
    return low+avg



import os

from setuptools import find_packages, setup

import versioneer


rootpath = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    return open(os.path.join(rootpath, *parts), 'r').read()


long_description = f'{read("README.rst")}\n{read("CHANGES.txt")}'


with open('requirements.txt') as f:
    tests_require = f.readlines()
install_requires = [t.strip() for t in tests_require]


setup(
    name='ioos_tools',
    version=versioneer.get_version(),
    packages=find_packages(),
    cmdclass=versioneer.get_cmdclass(),
    license=f'{read("LICENSE.txt")}',
    long_description=long_description,
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Development Status :: 4 - Beta'],
    description='Misc functions for IOOS notebooks',
    author=['Rich Signell', 'Filipe Fernandes'],
    author_email='ocefpaf@gmail.com',
    url='https://github.com/pyoceans/ioos_tools/releases',
    platforms='any',
    keywords=['oceanography', 'data analysis'],
    install_requires=install_requires,
    zip_safe=False,
)
